/*
 * Panasync Tools
 *
 * Copyright (C) 2001 Nuno Lopes
 * Copyright (C) 2000-2001 Carlos Baquero, Victor Fonte,
 * Paulo Sergio Almeida, Universidade do Minho
 *
 * panasync@gsd.di.uminho.pt - http://gsd.di.uminho.pt/panasync
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */


#include "panadefaults.h"
#include "system.h"
#include "pts.h"
#include "pds.h"


#ifndef PANACMDS_H
#define PANACMDS_H

/** Definition of debug/verbose level available for all modules/classes.
 * 
 * 0 - no verbose
 * 1 - verbose command errors only
 * 2 - verbose command errors and statements
 * 2 - verbose above plus pds errors
 */
extern int PANA_debug_level;

/** Consider potencial warnings as errors and abort operation.
 *
 * 0 - don't
 * 1 - consider
 */
extern int PANA_warning_to_error;


// PANASYNC COMMANDS


/** Initialize a base for use in the creation of panasync files.
 *
 * @param kind Char identifying kind of base.
 *  Possible values are 'u','s','r'.
 * @param id Unique identification for the base kind.
 * @param basefile String (\0 terminated) for base path.
 *  If null consider default base path that is /usr/share/panasync/base.
 * @return Pana error code. 0 if OK, code of error otherwise.
 */
int panainit(const char kind, const char * id, const char * basefile);


/** Create a new PANASYNC file. 
 *
 * @param filename Path for the new file.
 * @param contentsfilename Path for optional file that will suply the
 *  new file contents. If null, left new file unchanged or create an
 *  empty file (if it does not exist already).
 * @param basefilename Path for the file containing the base info that
 *  will be used for this new file. If null, use default base path instead.
 * @return Pana error code. 0 if OK, error code otherwise.
 */
int pananew(const char * filename, const char * contentsfilename,
				const char * basefilename);



// #### DUP section


/** Copies multiple sources to one target directory.
 *
 * It's a macro for panadirdup, which in turn is a macro for panadup.
 * It simply picks one source at a time and calls panadirdup with
 *  it and the target.
 *
 * @param args Number of strings stored in the array source.
 * @param source The array of strings (\0 terminated).
 * @param target The path of the target.
 * @return The result given by panadirdup.
 */
int panamultidup( int args, char ** source, char * target);



/** Copies source to target (directories).
 *
 * This function is a macro for panadup( sourcefile, targetfile).
 * It examines the arguments and calls panadup accordingly.
 *
 * @param source The path of the source file or directory.
 * @param target The path of the target file or directory.
 * @return The result of panadup or a PANA error code for 
 *  self-detected errors.
 */
int panadirdup( const char * source, const char * target);


/** Copies source file to target.
 *
 * By copy, in PANASYNC terms, is meant contents copy 
 * (like the 'copy' command) and forking the time-stamp.
 * This operation, forking the time-stamp, is necessary
 * for placing in 'time' the two files and be able to compare them.
 *
 * @param source The path to the source file.
 * @param target The path to the target file.
 * @return PANA error code. 0 if OK, otherwise an error code.
 */
int panadup( const char * source, const char * target);



// #### MOVE section


/** Moves multiple sources to one target directory.
 *
 * It's a macro for panadirmove, which in turn is a macro for panamove.
 * It simply picks one source at a time and calls panadirmove with
 *  it and the target.
 *
 * @param args Number of strings stored in the array source.
 * @param source The array of strings (\0 terminated).
 * @param target The path of the target.
 * @return The result given by panadirmove.
 */
int panamultimove( int args, char ** source, char * target);


				
/** Moves source to target (support for directories).
 *
 * This function is a macro for panamove( sourcefile, targetfile).
 * 
 * It examines the arguments and make the move procedure accordingly.
 * For regular files it just moves them, for pana-enabled files it moves
 * them and the meta-data binded to. 
 *
 * For directories, if the target directory already exists a new
 * sub-directory is created in the target (directory) and the source 
 * contents is moved into it. If the target does not exist, the contents of
 * source is moved into the newly created target directory.
 *
 * @param source The path of the source file or directory.
 * @param target The path of the target file or directory.
 * @return PANA error code. 0 if OK, otherwise error code.
 */
int panadirmove( const char * source, const char * target);



/** Move source file to target.
 *
 * Moving is creating a file, target with the
 *  contents of the source and removing the source.
 * With a PANASYNC file, the time-stamp must also be 'moved':
 *  erased from the source pds file and inserted into the target pds
 *  directory
 *
 * @param source The path to the source file.
 * @param target The path to the target file.
 * @return PANA error code. 0 if OK, otherwise an error code.
 */
int panamove( const char * source, const char * target);


/** Joins two files, resulting just one of them, the target.
 *
 * Joining two files is removing the source, placing the newer (or older)
 * contents in the target file and updating the time-stamp.
 *
 * If the two files are concurrent versions and if a contents file is
 *  suplied, the target will join the two time-stamps and the contents
 *  will be copied from it.
 *
 * @param source The path to the source file.
 * @param target The path to the second source and target file.
 * @param contents The path to an alternative contents file
 *  (may not be needed).
 * @param older Flag for reversing decision and storing the older file
 *  instead of the new one. 1 for reversing otherwise 0.
 * @return PANA error code. 0 if OK, otherwise an error code.
 */
int panajoin( const char * target, const char * source,
				const char * contents, int older);



/** Lists an array of paths.
 *
 * Is a macro for panalist.
 */
int panamultilist( int args, char ** path, int options, const char* compare);


/** Lists panafiles in a directory.
 *
 * @param pathname Directory or panafile to show info of...
 * @param options Options suported
 *  P_list_rec : recursive listing of all sub-dirs,
 *  P_list_long : display all filedata fields,
 *  P_list_frag : display only fragments for each file,
 *  P_list_cluster : group files by base.
 * @param comparefile File to be compared with every listed ones.
 * @return PANA_code_error.
 */
int panalist( const char *pathname, int options, const char * comparefile);

				

/** Make the translation between error codes and strings. 
 *
 * @param error_num The code of the panasync error.
 * @return The corresponding string if one applies.
 */
const char * panaerror( int error_num);

				
#endif  // PANACMDS_H

