/*
 * Panasync Tools
 *
 * Copyright (C) 2001 Nuno Lopes
 * Copyright (C) 2000-2001 Carlos Baquero, Victor Fonte,
 * Paulo Sergio Almeida, Universidade do Minho
 *
 * panasync@gsd.di.uminho.pt - http://gsd.di.uminho.pt/panasync
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */


#include "pds.h"


/** Creates a new empty PDS instance. */
pds::pds()
{
	// initialization of strings
	//  Not needed because object is not functional
	//  until opening a directory succeed.
	//
//	dirname = System::currentDirectory;
//	pdsname = PDS_filename;

	// initialization of flags
	//
	open = false;
	dirty = false;
//	om = read_only;

	dir.clear();
}

pds::~pds(){
}


/** Check if PANA directory status is consistent. */
// FIXME: not implemented properly
bool pds::checkDir( const string dirname, 
				const string pdsname = PDS_filename )
{
/*
	// create backup file name
	//
	string backup = dirname;
	backup += System::fileSystemSeparator;
	backup += pdsname;
	backup += PDS_backup_suffix;

	// check for existence of backup file
	//
	if ( System::exists( backup.c_str()) )
		// backup exists -> last operation didn't end correctly
		return false;
*/
	return true;
}

/** Open a panasync directory.
 *
 * @param dirname_p		The name of the directory from data is to be read.
 *		Should be a directory with at least read permission. If state 
 *		modifying operations are to be performed it should have also
 *		write permission.
 *
 * @param pdsfile_p		The name of the pds file in the directory.
 * 		Should be a regular file with read permission. Like the directory,
 * 		it should be writable if modifications are made in the directory.
 * 		If it doesn't exists, it is supposed that directory is empty of
 * 		PANASYNC files.
 *
 * @return	True if directory was open.
 */
bool pds::openDir( const string dirname_p,
				const string pdsname_p = PDS_filename )
{
	// check if directory is already opened
	//
	if ( open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:open: Trying to open an already opened directory."))
		return false;
	}

	// check if directory status is OK
	// if directory is not clean, open should fail!
	//
	if ( ! checkDir( dirname_p, pdsname_p))
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:open: checking status of directory failed."))
		return false;
	}

	// the PDS filename is dirname + '/' + pdsname
	//
	string file = dirname_p;
	file += System::fileSystemSeparator;
	file += pdsname_p;

	// clear mapping
	//
	dir.clear();
	
	// check if pds file exists
	//
	if ( System::exists( file.c_str()) )
	{
		// load data into object
		//
		read( file.c_str());
	}

	// initialize remaining values
	//
	dirname = dirname_p;
	// check if has filesystem separator at the end
	if ( dirname[ dirname.size()-1] != System::fileSystemSeparator)
		dirname += System::fileSystemSeparator;

	pdsname = pdsname_p;

	open = true;

	dirty = false;

	return true;
}

/** Closes a panasync directory.
 *
 * This method writes all data to the pds file if modifications were made
 *  on the directory. If the directory is empty, the pds file is
 *  removed.
 *
 * Requires permission to write on the pds file or in the directory,
 * 	if directory is dirty, i.e. modified.
 *
 * @return True, if information was successfuly writen to file. False,
 * 	if directory wasn't open or an error occurred in writing.
 */
bool pds::closeDir()
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:closedir: Trying to close without directory open."))
		return false;
	}

	// check if data is dirty, that is, modified
	//
	if ( dirty )
	{
		// build pds file name
		//
		string name = dirname;
		name += pdsname;

		// if data is not empty, write it to the file!
		if ( ! empty())
			write( name.c_str());

		// if directory is empty, remove pds file.
		// 
		// This removal is only done if directory was modified, allowing
		// possible read-only mode to not need write permission to directory
		//
		else
			System::remove( name.c_str());
	}

	// flags directory close
	//
	open = false;
	
	return true;
}


// internal data operations


/** Is directory empty, no panasync files. */
bool pds::empty() const
{
	// check if directory is open
	//
	if ( !open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:empty: Directory was not opened."))
		return false;
	}

	return dir.empty();
}

/** Clear all data. */
void pds::clear()
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:clear: Directory was not open."))
		return;
	}

	// flags dirty
	//
	dirty = true;

	dir.clear();
}
	
/** Is filename a PANASYNC file in this directory. */
bool pds::hasFile( const char * filename) const
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:hasFile: Directory was not opened."))
		return false;
	}
	
	return (dir.find( filename) != dir.end());
}

/** Has file been modified since last update/insert. */
bool pds::isFileModified( const char * filename)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:isFileModified: Directory was not opened."))
		return false;
	}

	hash_t hash;
	string file = dirname;
	file += filename;
	
	// calculate md5 hash for filename
	digest( file.c_str(), hash);

	return (memcmp( dir[filename].md5, hash, MD5BUFSIZE) != 0);
}


// INSERT section


/** Inserts a new PANA file with given base.
 * 
 * Requires read access to the file because md5 is evaluated from it. 
 *
 * @return True if file was inserted.
 */
bool pds::insertFile( const char * filename, const pts::base & base)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:insertFile: Directory was not opened."))
		return false;
	}

	// calculate complete file name path
	//
	string file = dirname;
	file += filename;

	// calculate md5 hash for filename
	//
	hash_t md5;
	digest( file.c_str(), md5);

	return insertFile( filename, md5, base );
}

/** Inserts a new PANA file with given time-stamp.
 * 
 * Requires read access to the file because md5 is evaluated from it. 
 *
 * @return True if file was inserted.
 */
bool pds::insertFile( const char * filename, const pts & stamp)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:insertFile: Directory was not opened."))
		return false;
	}

	// calculate complete file name path
	//
	string file = dirname;
	file += filename;

	// calculate md5 hash for filename
	//
	hash_t md5;
	digest( file.c_str(), md5);

	return insertFile( filename, md5, stamp);
}

/** Inserts a new PANA file with given hash and base.
 * 
 * @return True if file was inserted.
 */
bool pds::insertFile( const char * filename, const hash_t md5,
				const pts::base & base)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:insertFile: Directory was not opened."))
		return false;
	}

	// creates a time-stamp from given base
	//
	pts stamp(base);

	return insertFile( filename, md5, stamp );
}

/** Inserts a new PANA file with given hash and time-stamp.
 * 
 * @return True if file was inserted.
 */
bool pds::insertFile( const char * filename, const hash_t md5,
				const pts & stamp)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:insertFile: Directory was not opened."))
		return false;
	}

	// check if file already exists,
	//
	if ( hasFile( filename))
	{
		ERROR( PANA_debug_pds, gettext("pds:insertFile(name,md5,hash): File already exists."))
		return false;
	}

	// insert the pair name,filedata in dir
	//
	dir[filename]= filedata( md5, stamp);

	// flags dirty
	//
	dirty = true;

	return true;
}

/** Inserts a new PANA file with associated filedata structure.
 * 
 * @return True if file was inserted.
 */
bool pds::insertFile( const char * filename, const filedata & fdata)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		ERROR( PANA_debug_pds, gettext("pds:insertFile: Directory was not opened."))
		return false;
	}

	// check if file already exists
	//
	if ( hasFile( filename))
	{
		ERROR( PANA_debug_pds, gettext("pds:insertFile(name,filedata): File already exists."))
		return false;
	}

	// insert filedata structure in dir
	//
//	dir[filename]= filedata(fdata);
	dir[filename]= fdata;

	// flags dirty
	//
	dirty = true;
	
	return true;
}

/** Updates information for file name if it was modified.
 *
 * If the file was changed, it's contents will give a different md5 hash.
 * If it was the case, the filedata info associated to the file
 *  will be updated.
 * 
 * Requires read permission to file because md5 hash is evaluated from it.
 * 
 * @return True if file data was updated in the directory or no error found.
 *  False otherwise.
 */
bool pds::updateFile( const char * filename)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		ERROR( PANA_debug_pds, gettext("pds:updateFile: Directory was not opened."))
		return false;
	}

	// check if file not exists
	//
	if ( ! hasFile( filename))
	{
		ERROR( PANA_debug_pds, gettext("pds:updateFile: File does not exist."))
		return false;
	}

	// complete file name path
	//
	string file = dirname;
	file += filename;

	// calculate new md5 hash for filename
	//
	hash_t hash;
 	digest( file.c_str(), hash);

	// if file has NOT been modified ...
	// return true because no error was found and file is already up to date
	//
	if (memcmp( dir[filename].md5, hash, MD5BUFSIZE) == 0)
		return true;

	return updateFileHash( filename, hash);
}

/** Updates md5 hash and time-stamp information for this filename
 *  in the directory.
 *
 * @return True if file data was updated. False if file does not exist.
 */
bool pds::updateFileHash( const char * filename, const hash_t md5)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		ERROR( PANA_debug_pds, gettext("pds:updateFileHash: Directory was not opened."))
		return false;
	}

	// check if file not exist
	//
	if ( ! hasFile( filename))
	{
		ERROR( PANA_debug_pds, gettext("pds:updateFileHash: File does not exist."))
		return false;
	}

	// update md5 hash and set time-stamp dirty
	//
	memcpy( dir[filename].md5, md5, MD5BUFSIZE);
	dir[filename].stamp.set_dirty();

	// flags dirty
	//
	dirty = true;

	return true;
}

/** Remove PANASYNC file from directory.
 *
 * Remember not to confuse this directory -> pds data, with the real one.
 * Removing a file in here just makes it a non-PANASYNC file. It does NOT
 * remove it from disk.
 * 
 * @return True if file was removed from direstory data.
 */
bool pds::removeFile( const char * filename)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		ERROR( PANA_debug_pds, gettext("pds:removeFile: Directory was not opened."))
		return false;
	}

	// check if file does not exist
	//
	if ( ! hasFile( filename))
	{
		ERROR( PANA_debug_pds, gettext("pds:removeFile: File does not exist."))
		return false;
	}

	// remove entry
	//
	dir.erase( filename);

	// flags dirty
	//
	dirty = true;

	return true;
}


/** Retrieves filedata information of a file (copy). */
pds::filedata pds::getFileData( const char * filename)
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:getFileData: Directory was not opened."))
		return filedata();
	}

	// check if file does not exist
	//
	if ( ! hasFile(filename))
	{
		ERROR( PANA_debug_pds, gettext("pds:getFileData: File does not exist."))
		return filedata();
	}

	return filedata( dir[filename]);
}

/** Forks the time-stamp of a file.
 *
 * @return the new time-stamp created by forking the other.
 */
pts pds::forkFileStamp( const char * filename )
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:forkFileStamp: Directory was not opened."))
		return pts();
	}

	// check if file does not exist
	//
	if ( ! hasFile(filename))
	{
		ERROR( PANA_debug_pds, gettext("pds:forkFileStamp: File does not exist."))
		return pts();
	}

	// set flag dirty
	//
	dirty = true;

	// fork the time-stamp
	//
	return dir[filename].stamp.fork();
}

/** Forks a file time-stamp returning the new filedata created.
 *
 * @return A new filedata containing the same md5 hash and
 * 	the new time-stamp created by forking the other.
 */
pds::filedata pds::forkFile( const char * filename )
{
	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:forkFile: Directory was not opened."))
		return filedata();
	}

	// check if file does not exist
	//
	if ( ! hasFile(filename))
	{
		ERROR( PANA_debug_pds, gettext("pds:forkFile: File does not exist."))
		return filedata();
	}

	// set flag dirty
	//
	dirty = true;

	// fork the time-stamp and return a new filedata struct with it
	//
	return filedata( dir[filename].md5, dir[filename].stamp.fork());
}

/** Lists file names in the PANASYNC directory. */
vector<string> pds::listFileNames()
{
	// return value declaration
	//
	vector<string> ret;

	// check if directory is not opened
	//
	if ( ! open )
	{
		// if debug level is equal or higher than this class level
		//
		ERROR( PANA_debug_pds, gettext("pds:listFileNames: Directory was not opened."))
		return ret;
	}

	// find all names in the map
	//
	for( map_i i= dir.begin(); i != dir.end() ; i++)
		ret.push_back( i->first);

	return ret;
}




// auxiliary functions


/** Digest a file creating a hash. */
void pds::digest( const char * filename, hash_t hash){

	FILE * fp = fopen( filename, "r");
	md5_stream(fp, hash);
	fclose(fp);
}


/** Reads data from a PDS file to the object data structures.
 *
 * @param filename The file name where the data is to be read.
 */
bool pds::read( const char * filename ){

	// Establish externalization environment
	//
	unsigned char fmb[PTBUFSIZE];	// memory buffer
 
	// Establish compression environment
	//
	Bytef cbuf[PTBUFSIZE];	// Compression buffer
	unsigned long csize = PTBUFSIZE;	// Compression buffer size
 
	// Initialize input stream
	//
	ifstream fp( filename );
/*
	// declaration of version buffer
	//
	char vbuf[ strlen(version.c_str())+1 ];
	
	// Read version from serialized file
	//
	fp.read(vbuf, sizeof(vbuf));
	if ( version.compare(vbuf) )
		return false;
*/
	// Read buffer from file 
	//
//	fp.read(fmb, sizeof(fmb));
	fp.read(cbuf, sizeof(cbuf));
	fp.close();

	// Uncompress externatization buffer
	//
	if (uncompress((Bytef *)fmb, &csize, cbuf, PTBUFSIZE) != Z_OK)
		return false;

	// Revert externalization of Panasync's data structures
	//
	mem_buffer xmb(fmb, sizeof(fmb));  // XTL memory buffer object
	GIOP_format<mem_buffer> gfmb(xmb);
	obj_input<GIOP_format<mem_buffer> > is(gfmb);

	// clear all mapping data
	dir.clear();
	// read the data
	is.simple( dir);
	
	return true;
}


/** Writes the object data structures into a PDS file.
 *
 * @param filename The file name where the data is to be stored.
 */
void pds::write( const char * filename ){

	// Establish externalization environment
	//
	unsigned char	fmb[PTBUFSIZE];		// memory buffer
	mem_buffer	xmb(fmb, sizeof(fmb));	// XTL memory buffer object
	GIOP_format<mem_buffer> gfmb(xmb);
	obj_output<GIOP_format<mem_buffer> > os(gfmb);

	// Establish compression environment
	//
	Bytef		cbuf[PTBUFSIZE];	// Compression buffer
	unsigned long	csize = PTBUFSIZE;	// Compression buffer size

	// Externalize Panasync's data strutures
	//
	os.simple(dir);

	// Compress externalization buffer
	// 
	// 	Compress levels:
	// 		Z_BEST_COMPRESSION,
	// 		Z_BEST_SPEED,
	// 		Z_DEFAULT_COMPRESSION
	//
	assert(compress2(cbuf, &csize, (Bytef *)xmb.data(), 
                  xmb.size(), Z_BEST_COMPRESSION) == Z_OK);

	// Write buffer to file	
	//
	ofstream fp( filename);
//	fp.write(xmb.data(), xmb.size());
	fp.write(cbuf, csize);
	fp.close();

	// return true;
}




// definition of friend output stream
//

ostream & operator< (ostream & out, const pds::hash_t & hash)
{
	char s[3];
	// for every byte in the hash
	// print it using hex notation
	//
//	for (unsigned int i=0; i<MD5BUFSIZE; i++){
	for (unsigned int i=0; i<sizeof(hash); i++){
		sprintf( s, "%02X", hash[i]);
		out << s ;
	}
	return out;
}

ostream & operator << (ostream & out, const pds::map_t & dir)
{
	// format heading
	out << ":name\t\t{md5_hash},{time-stamp=b kind.id.counter;frag 1:2}" << endl;

	// for each element
	//
	for (pds::map_t::const_iterator i=dir.begin(); i!=dir.end(); i++){
		out << i->first.c_str()
				<< "\t\t{";
		// special output for hash_t:sequence of bytes
		out < i->second.md5;
		out << "},"
				<< i->second.stamp
				<< endl;
	}
	return out;
}

ostream & operator << (ostream & out, const pds & pds)
{
	// version
	out << pds.version.c_str() << endl ;
	// directory name
	out << "- Directory name: " << pds.dirname.c_str() << endl ;
	// pds file name if not default
	if ( pds.pdsname != PDS_filename )
		out << "- PDS file name: " << pds.pdsname.c_str() << endl ;
	// warning if dirty
	if ( pds.dirty )
		out << "[modified]";
	// list of files in diretory
	out << "- List of files: " << endl ;
	// internal listing
	out << pds.dir;
	out << "--end--" << endl;

	return out;
}


// version of pds data structure (static)
const string pds::version = string("PDS v1.0");


