/*
 * Panasync Tools
 *
 * Copyright (C) 2001 Nuno Lopes
 * Copyright (C) 2000-2001 Carlos Baquero, Victor Fonte,
 * Paulo Sergio Almeida, Universidade do Minho
 *
 * panasync@gsd.di.uminho.pt - http://gsd.di.uminho.pt/panasync
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */


#include <stdio.h> // fopen, fclose
#include <zlib.h>	// compression lib
#include <fstream.h>	// {i,o}fstream
#include <string>
#include <stl.h>
#include <libintl.h> // gettext


extern "C" {
#include "md5.h" // md5 digests
}
#include "xtl.h"	// XTL


#include "panadefaults.h"
#include "pts.h"
#include "system.h"


#ifndef PDS_H
#define PDS_H


// constant data values

// size of md5 hash (bytes)
const unsigned MD5BUFSIZE	= 16;

// default file names
const string PDS_filename = PTDEF_PDS_FILENAME;
const string PDS_backup_suffix = ".backup";
const string PDS_log_suffix = ".oplog";



/** Panasync Directory Structure Class.
 *
 * This class is responsible for managing a set of pana-enabled files in
 *  a directory.
 * 
 * @version 1.0
 */
class pds{

public: 

	/** Creates a new empty PDS instance. */
	pds();
	
	~pds();


	// public structures

		
	/// Definition of hash type.
	typedef unsigned char hash_t[MD5BUFSIZE];


	/** Definition of filedata structure (XTL enabled).
	 * 
	 * This structure holds the information needed for each
	 *  PANA-enabled file.
	 *
	 * [struct] filedata
	 */
	struct	filedata
	{
	
	/// md5 hash for detecting modifications in file
	hash_t md5;
	/// time-stamp for controling version evolutions of the file
	pts stamp;

	filedata(){}

	/** Constructor for filedata struct.
	 * 
	 * This constructor takes his arguments and place them in the
	 *  structure fields.
	 * 
	 * @param srcmd5 MD5 hash
	 * @param srcstamp pts instance
	 */
	filedata( const hash_t srcmd5, const pts & srcstamp) : stamp(srcstamp)
	{ memcpy( md5, srcmd5, sizeof(md5));
	}
	
	/** Constructor for filedata struct.
	 *
	 * This constructor takes his arguments and place them in the
	 *  structure fields.
	 * 
	 * @param srcmd5 MD5 hash
	 * @param srckind base kind
	 * @param srcid base id
	 * @param srccounter base counter
	 */
	filedata(const hash_t srcmd5, string srckind, string srcid,
			string srccounter) : stamp(srckind, srcid, srccounter)
	{ memcpy( md5, srcmd5, sizeof(md5));
	}

	template <class Stream>
	void composite (Stream & stream){
		stream.vector( md5, sizeof(md5)).simple(stamp);
	}
	}; // end of pds::filedata

	
	/** Definition of possible open modes for a directory.
	 *
	 * [enum] open_mode
	 * FIXME: currently not used
	 */
	enum open_mode {
		/** Opens directory only for reading only,
		 *   state changing operations are not allowed,
		 *   no write operations are performed on the directory.
		 */
		read_only,
		/** Opens directory for normal operations,
		 *   without backup/recover possibility.
		 */
		read_write,
		/** Opens directory for normal operations,
		 *   with the possibility to recover from a crash using panacheck.
		 */
		rw_log,
		/** Opens directory for check and fix operations. */
		check
	}; // end of pds::open_mode


	/// Mapping between string and filedata types.
	typedef map<string, filedata> map_t;

	// Iterator for the mapping.
	typedef map_t::iterator map_i;


	// public methods


	/** Check if PANA directory status is consistent.
	 * FIXME: not properly implemented
	 */
	static bool checkDir( const string dirname,
			const string pdsname = PDS_filename);

	/** Open a panasync directory.
	 *
	 * @param dirname_p		The name of the directory from data is to be read.
	 *		Should be a directory with at least read permission. If state 
	 *		modifying operations are to be performed it should have also
	 *		write permission.
	 *
	 * @param pdsfile_p		The name of the pds file in the directory.
	 * 		Should be a regular file with read permission. Like the directory,
	 * 		it should be writable if modifications are made in the directory.
	 * 		If it doesn't exists, it is supposed that directory is empty of
	 * 		PANASYNC files.
	 *
	 * @return	True if directory was open.
	 */
	bool openDir( const string dirname_p,
				const string pdsname_p = PDS_filename );


	/* Open a panasync directory with crash recover capability.
	 * FIXME: not implemented yet
	bool opendir( const string dirname, open_mode mode,
			const string pdsname = PDS_filename );
	 */

	/** Closes a panasync directory.
	 *
	 * This method writes all data to the pds file if modifications were made
	 *  on the directory. If the directory is empty, the pds file is
	 *  removed.
	 *
	 * Requires permission to write on the pds file or in the directory,
	 * 	if directory is dirty, i.e. modified.
	 *
	 * @return True, if information was successfuly writen to file. False,
	 * 	if directory wasn't open or an error occurred in writing.
	 */
	bool closeDir();


	/** Is directory empty, no panasync files. */
	bool empty() const;

	/** Clear all data. */
	void clear();


	/** Is file a PANASYNC file in this directory? */
	bool hasFile( const char * filename) const;

	/** Inserts a new PANA file with given base.
	 * 
	 * Requires read access to the file because md5 is evaluated from it. 
	 *
	 * @return True if file was inserted.
	 */	
	bool insertFile( const char * filename, const pts::base & base);

	/** Inserts a new PANA file with given time-stamp.
	 * 
	 * Requires read access to the file because md5 is evaluated from it. 
	 *
	 * @return True if file was inserted.
	 */	
	bool insertFile( const char * filename, const pts & stamp);

	/** Inserts a new PANA file with given hash and base.
	 * 
	 * @return True if file was inserted.
	 */
	bool insertFile( const char * filename, const hash_t md5,
			const pts::base & base);

	/** Inserts a new PANA file with given hash and time-stamp.
	 * 
	 * @return True if file was inserted.
	 */	
	bool insertFile( const char * filename, const hash_t md5,
			const pts & stamp);
	
	/** Inserts a new PANA file with associated filedata structure.
	 * 
	 * @return True if file was inserted.
	 */
	bool insertFile( const char * filename, const filedata & fdata);


	/** Has file been modified since last update/insert. */
	bool isFileModified( const char * filename);

	/** Updates information for file name if it was modified.
	 *
	 * If the file was changed, it's contents will give a different md5 hash.
	 * If it was the case, the filedata info associated to the file
	 *  will be updated.
	 * 
	 * Requires read permission to file because md5 hash is evaluated from it.
	 * 
	 * @return True if file data was updated in the directory.
	 */
	bool updateFile( const char * filename);

	/** Updates md5 hash and time-stamp information for this filename
	 *  in the directory.
	 *
	 * @return True if file data was updated. False if file does not exist.
	 */
	bool updateFileHash( const char * filename, const hash_t md5);


	/** Remove PANASYNC file from directory.
	 *
	 * Remember not to confuse this directory -> pds data, with the real one.
	 * Removing a file in here just makes it a non-PANASYNC file. It does NOT
	 * remove it from disk.
	 * 
	 * @return True if file was removed from directory data.
	 */
	bool removeFile( const char * filename);


	/** Retrieves filedata information of a file (copy). */
	filedata getFileData( const char * filename);

	/** Forks the time-stamp of a file.
	 *
	 * @return the new time-stamp created by forking the other.
	 */
	pts forkFileStamp( const char * filename);

	/** Forks a file time-stamp returning the new filedata created.
	 *
	 * @return A new filedata containing the same md5 hash and
	 * 	the new time-stamp created by forking the other.
	 */
	filedata forkFile( const char * filename );


	/** Lists file names in the directory. */
	vector<string> listFileNames();


	/** Digest a file creating a hash. */
	static void digest( const char * filename, hash_t hash);


private:

	
// methods

	/** Reads data from a PDS file to the object data structures.
	 *
	 * @param filename The file name where the data is to be read.
	 */
	bool read( const char * filename );

	/** Writes the object data structures into a PDS file.
	 *
	 * @param filename The file name where the data is to be stored.
	 */
	void write( const char * filename );

// variables
	
	/** 
	 * Mapping between file names and filedata structures for the
	 * PANASYNC files contained in a directory.
	 */
	map_t dir;

	/** Is a directory open. */
	bool open;

	/** Has directory been modified by some state-modifying operation. */
	bool dirty;

	/** Current directory pathname. */
	string dirname;

	/** File name of pds data structure. */
	string pdsname;

	/* Open mode requested for current directory. */
	// FIXME: not currently supported
//	open_mode om;


	/** Version of pds data structures (stored in the pds file). */
	const static string version;


	//	friends
	//
	friend ostream & operator<< (ostream & out, const pds & pds);
	friend ostream & operator<< (ostream & out, const pds::map_t & dir);
	friend ostream & operator< (ostream & out, const pds::hash_t & hash);


};

#endif  // PDS_H

