/*
 * Panasync
 *
 * Copyright (C) 2000 Carlos Baquero, Victor Fonte,
 * Paulo Sergio Almeida, Universidade do Minho
 *
 * panasync@gsd.di.uminho.pt - http://gsd.di.uminho.pt/panasync
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */


#include <stdio.h> // sprintf
#include <string>
#include <fstream.h>	// {i,o}fstream
#include <stl.h>


#include "xtl.h"	// XTL


#include "panadefaults.h"


// include fake function
// 
//#define PHONY_F

// generate full output when <<'ing
//
#define PTS_FULL_OUTPUT


#ifndef PTS_H
#define PTS_H

/** Panasync Time-Stamp Class.
 * 
 * This class is responsible for managing the time-stamp information for
 *  each pana-enabled file.
 *
 * @version 1.0
 */
class pts
{
	public:


	// public structures


	/** Definition of time-stamp base type. */
	struct base {
	/** Specifies the kind/type of the base.
	 * 
	 * Possible values are: (universaly) unique, string and random
	 */
	string	kind;
	/// identifies a unique base for each kind
	string	id;
	/** distinguish the stamps that where created with the same pair
	 *   (kind, id).
	 */
	string	counter;

	/// Increments counter field of base.
	void inc(){
		// increment counter field
		//
		int value = strtol( counter.c_str(), NULL, 10);
		value++;
		char word[15];
		sprintf( word, "%d", value);
		counter = word;
	}

	/** Reads from file the base data structure (using XTL).*/
	void read( const char * filename){

		// Establish externalization environment
		//
		char buf[PTBUFSIZE];		// Change into a XTL autobuffer!
		ifstream basefile( filename);
		basefile.read(buf, sizeof(buf));
		basefile.close();

		// Reverse externalization
		//
		mem_buffer mb(buf, sizeof(buf));
		GIOP_format<mem_buffer> gfmb(mb);	
		obj_input<GIOP_format<mem_buffer> > is(gfmb);
		is.simple(kind).simple(id).simple(counter);
	}

	/** Writes to file the base data structure (using XTL).*/
	void write( const char * filename){

		// Establish externalization environment
		//
		char buf[PTBUFSIZE];		// Change into a XTL autobuffer!
		mem_buffer	mb(buf, sizeof(buf));
		GIOP_format<mem_buffer> gfmb(mb);	
		obj_output<GIOP_format<mem_buffer> > os(gfmb);
	
		// Externalize Panasync's base
		//
		os.simple(kind).simple(id).simple(counter);
	
		// Write externalized Panasync base
		//
		ofstream	basefile( filename );
		basefile.write(mb.data(), mb.size());
		basefile.close();
	}
	};	// end pts::base


	// public constructors


	pts();

	/** Creates a new time-stamp with the specified values for the base.
	 *
	 * @param kind The kind for the base used.
	 * @param id The identification for the base used.
	 * @param counter The counter for the base used.
	 */
	pts(string kind, string id, string counter);

	/** Creates a new identical time-stamp from another. */
	pts(const pts & source);

	/** Creates a new time-stamp from a base. */
	pts(const base & b);

	~pts();



	// public methods


	/** Sets the time-stamp as modified in the present state. 
	 * This implies coping the second fragment to the first.
	 */
	void	set_dirty();
	
	/** Forks the time-stamp.
	 * Creates a new time-stamp with a forked version
	 * and updates its version also.
	 */
	pts	fork();
	
	/** Joins two time-stamps.
	 * 
	 * This operation join two different time-stamps into one.
	 */
	pts &	join(const pts & relative);
	
	/** Check if the time-stamp dominates (is more ??? than) the other. */
	bool	dominates(const pts & relative) const;
	
	/** Check if the time-stamp is equivalent to the other. */
	bool	equivalent(const pts & relative) const;
	
	/** Check if the time-stamp and the other are uncomparable.
	 *
	 * Being uncomparable means that they have different bases. 
	 */
	bool	uncomparable(const pts & relative) const;

	template <class Stream> void composite(Stream& stream);

#ifdef PHONY_F
	// fake function
	void phony();
#endif // PHONY_F

	/** Get strig'alized base of instance. */
	string getBase();
	
	/// Part is a sequence (vector) of boolean digits (bool).
	typedef vector<bool> part;
	
	/// Fragment is a collection (set) of parts.
	typedef set<part> frag;


private:

	//	declaration of the fragment objects
	
	/** time-stamp base.
	 * 
	 * Identifies the base values for this time-stamp
	 */
	base	bs;
	
	/** time-stamp first fragment.
	 * 
	 * Shows the last modification 'time' of the time-stamp.
	 */
	frag	ff;
	
	/** time-stamp second fragment.
	 *
	 * Shows the present 'time' of the time-stamp.
	 */
	frag	sf;

	//	private methods
	//
	void	distribute(frag & ds, frag & fs1, frag & fs2, frag & fs3);

	//	friends
	//
	friend ostream& operator<< (ostream& out, const pts& pts);

	//	debug
	//
#ifdef	__panasync_debug_causal_history
	set<int> ch;
#endif  // __panasync_debug_causal_history
};

#endif  // PTS_H
