/*
 * Panasync Tools
 *
 * Copyright (C) 2001 Nuno Lopes
 * Copyright (C) 2000-2001 Carlos Baquero, Victor Fonte,
 * Paulo Sergio Almeida, Universidade do Minho
 *
 * panasync@gsd.di.uminho.pt - http://gsd.di.uminho.pt/panasync
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */


#include <unistd.h>	// access, stat
#include <fcntl.h>	// creat
#include <sys/types.h>	// creat, opendir, readdir, closedir
#include <sys/stat.h>	// stat
#include <sys/times.h> // times
#include <string.h>	// strrchr
#include <dirent.h> // opendir, readdir, closedir


#include "system.h"


/** Check if pathname exists in filesystem */
bool System::exists( const char * pathname){

	return (access( pathname, F_OK)==0);
}

/** Check if pathname is a directory */
bool System::isDir( const char * pathname){

	struct stat fstat;
	return (stat( pathname, &fstat)==0) && S_ISDIR(fstat.st_mode);
}

/** Check if pathname is regular file.
 * 
 * It considers symbolic links as regular files (is this a bug?).
 */
bool System::isFile( const char * pathname){

	struct stat fstat;
	return (stat( pathname, &fstat)==0) && S_ISREG(fstat.st_mode);
}

/** Check if pathname is readable */
bool System::isReadable( const char * pathname){

	return (access( pathname, R_OK)==0);
}

/** Check if pathname is writable */
bool System::isWritable( const char * pathname){

	return (access( pathname, W_OK)==0);
}

/** Check if pathname is readable and writable */
bool System::isReadWriteAble( const char * pathname){

	return (access( pathname, R_OK + W_OK)==0);
}

/** Check if pathname is readable and writable */
bool System::isEmpty( const char * pathname){

	struct stat fstat;
	return (stat( pathname, &fstat)==0) && (fstat.st_size == 0);
}

/** Check if pathnames are different files acording to filesystem rules */
// FIXME: add string compare on names, is it necessary ?
bool System::isDifferent( const char * pathname1, const char * pathname2){

	struct stat fstat1, fstat2;
	return !( (stat( pathname1, &fstat1)==0) && (stat( pathname2, &fstat2)==0)
		&& (fstat1.st_ino == fstat2.st_ino) );
}

/** Creates regular file with pathname on filesystem */
bool System::create( const char * pathname)
{
	bool ret;
	int fd= creat( pathname, 0644);
	ret = (fd>-1);
	if (ret)
		close(fd);
	return ret;
}

/** Creates a directory with specified pathname. */
bool System::createDir( const char * pathname)
{
	return mkdir( pathname, 0755) > -1;
}

/** Copy source contents to target */
bool System::copy( const char * source_pathname,
			const char * target_pathname){
	// values for number of bytes readed/writed
	int n, m;
	// buffer for transfer operation
	char buffer[1024];
	
	// file descriptor for source
	int ifs = open( source_pathname, O_RDONLY);
	if ( ifs ==-1 )
		return false;
	// file descriptor for target
	int ofs = creat( target_pathname, 0622);
	if ( ofs ==-1 ){
		close (ifs);
		return false;
	}

	while ( (n = read(ifs, buffer, sizeof(buffer))) > 0){
		// assert(write (ofs, buffer, n) == n);
		m = write(ofs, buffer, n);
		// error during write
		if ( m < n) break;
	}

	close (ifs);
	close (ofs);

	// on error, remove second file
	if ( n<0 || m<0 ){
		remove( target_pathname);
		return false;
	}

	return true;
}

/** Removes file on the filesystem. */
bool System::remove( const char * pathname){

	return ( unlink( pathname) == 0);
}

/** Removes directory on the filesystem. */
bool System::removeDir( const char * pathname){

	return ( rmdir( pathname) == 0);
}


// string manipulation for path names

/** Generate a complete random string with the specified length. */
char * System::generateString(char * str, int length)
{
	// check if string is null
	//
	if ( str == NULL)
		return str;

	// initialize random seed
	//
	srandom( times(NULL));

	// fill string with values rounded for characters
	//
	for( int i=0; i < length-1; i++)
		str[i] = random() % 24 + 'a';
	
	// end string with char 0
	// 
	str[length-1] ='\0';

	return str;
}

/** Removes the FileSystemSeparator from the end of the string. */
char * System::removeLastSeparator( char * pathname)
{
	int length = strlen( pathname);
	while ( length>0
			 && pathname[length-1]== System::fileSystemSeparator )
	{
		pathname[--length]= '\0';
	}
	return pathname;
}


/** Extracts from pathname the directory path */
char * System::getDirName( const char * pathname, char * dirname,
		int maxlength){

	// previous name - panacmds.cc:getbasedir()

	// clear string
	if ( dirname) dirname[0]=0;

	// finds last occurrence of fileSystemSeparator in pathname
	char * p = strrchr(pathname, fileSystemSeparator);

	// if separator exists
	// copy first half into dirname
	if (p != NULL)
	{
		if ( maxlength >= ( p - pathname + 2) )
			maxlength = p - pathname + 1;
		else
			maxlength--;
		
		//strncpy(dirname, pathname, p - pathname + 1);
		strncpy(dirname, pathname, maxlength);
		dirname[maxlength]=0;
	}
	// if does not exist return './'
	else
		if ( maxlength > 2)
			strcpy(dirname, "./");

	return dirname;
}


/** Extracts the file name from pathname */
char * System::getFileName( const char * pathname, char * filename,
				int maxlength){

	// previous location - panacmds.cc:getbasename()

	// clear string
	if (filename) filename[0]=0;

	// finds last occurrence of '/'
	char * p = strrchr(pathname, fileSystemSeparator);

	// copies into filename the last piece of pathname
	// or all of it if did not found a separator
	strncpy(filename, (p == NULL ? pathname : p+1 ), maxlength-1 );
	filename[maxlength-1]=0;
	
	return filename;
}

/** List of files (not directories) in the specified directory */
vector<string> System::listFiles( const char * dirname){

	// creation of return value
	vector<string> * ret = new vector<string>();

	struct dirent * s;
	DIR * d = opendir( dirname);
	string entryname;

	if ( d != NULL){
		while ( (s=readdir(d)) != NULL){

			entryname = dirname;
			entryname += System::fileSystemSeparator;
			entryname += s->d_name;

			if ( System::isFile( entryname.c_str()) )
				ret->push_back( string(s->d_name));
		}
	}
	return *ret;
}

/** List of sub-directories in the specified directory */
vector<string> System::listDirs( const char * dirname){

	// creation of return value
	vector<string> * ret = new vector<string>();

	struct dirent * s;
	DIR * d = opendir( dirname);
	string entryname;

	if ( d != NULL){
		while ( (s=readdir(d)) != NULL){

			// directory name (without path)
			string name = s->d_name;

			// complete directory pathname
			entryname = dirname;
			entryname += System::fileSystemSeparator;
			entryname += s->d_name;

			if ( System::isDir( entryname.c_str())
				&& (name.compare("..") != 0)
				&& (name.compare(".")  != 0) )
//				ret->push_back( entryname);
				ret->push_back( name);
		}
	}
	return *ret;
}


/** Defines the current directory pathname. */
const char * System::currentDirectory = "./";

