/*
 * Panasync Tools
 *
 * Copyright (C) 2001 Nuno Lopes
 * Copyright (C) 2000-2001 Carlos Baquero, Victor Fonte,
 * Paulo Sergio Almeida, Universidade do Minho
 *
 * panasync@gsd.di.uminho.pt - http://gsd.di.uminho.pt/panasync
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */


#include <string>
#include <stl.h>


#ifndef SYSTEM_H
#define SYSTEM_H


/** System Functions.
 * 
 * Suplies the operating system functions needed for the aplication.
 * It's system specific and should be mapped to the target OS.
 * 
 * Remember that every function/method is static!
 * 
 * @version 1.0
 */
class System {

public: 

 /** Check if pathname exists in the filesystem.
  */
  static bool exists( const char * pathname);

 /** Check if pathname is writable.
	*
	* If path is a link, the test aplies to the file referred to.
  */
  static bool isWritable( const char * pathname);

 /** Check if pathname is readable.
	*
	* If path is a link, the test aplies to the file referred to.
  */
  static bool isReadable( const char * pathname);

 /** Check if pathname is readable and writable.
	*
	* If path is a link, the test aplies to the file referred to.
  */
  static bool isReadWriteAble( const char * pathname);

 /** Check if pathname is a regular file.
  */
  static bool isFile( const char * pathname);

 /** Check if pathname is a directory.
  */
  static bool isDir( const char * pathname);

 /** Check if pathname is readable and writable */
	static bool isEmpty( const char * pathname);

 /** Check if pathnames are different files acording to filesystem rules.
	*
  * Some systems support links refering to another file from the first.
  * Some operations need to ensure that the two files are not the same.
  * 
  * The semantics of this operation is to verify if there is a link
  *  from one to another.
  *
  * This operation does not check/uses the contents of the files!
  * Instead, (in Unix-compatible systems) it uses the i-node of each file.
  *
  * @param pathname1_2 The two filenames to be compared.
  * @return True if they are different files.
  */
  static bool isDifferent( const char * pathname1, const char * pathname2);

 /** Create a regular file in the filesystem.
  *
  * The content of the file newly created is empty.
  *
	* @param pathname The name of the file.
  * @return True if operation was successfull.
  */
  static bool create( const char * pathname);

 /** Creates a directory with specified pathname.
	*
	* @param pathname The name of the directory.
	* @return True if operation was successfull.
	*/
	static bool createDir( const char * pathname);

 /** Copy source contents to target.
  *
  * This function has the same effect of making a 'copy' command on the shell.
  *
  * @param source_pathname The source file name.
	* @param target_pathname The target file name.
  * @return True if operation was successfull.
  */
  static bool copy( const char * source_pathname,
			const char * target_pathname);

 /** Removes file on the filesystem.
  *
	* @param pathname The name of the file.
  * @return True if operation was successfull.
  */
  static bool remove( const char * pathname);

 /** Removes a directory on the filesystem.
  *
	* @param pathname The name of the directory.
  * @return True if operation was successfull.
  */
  static bool removeDir( const char * pathname);

 /** Generate a complete random string with the specified length. 
	*
	* @param str The array to where characters will be written.
	* @param length The characters number of the string.
	* @return The string.
  */
	static char * generateString(char * str, int length);

 /** Removes the FileSystemSeparator from the end of the string.
	*
	* If there is no separator at the end, return the same string.
	*
	* @param pathname The string containing the path.
	* @return The same reference to the string.
	*/
	static char * removeLastSeparator( char * pathname);

 /** Extracts the file name component from a pathname.
  *
  * @return File name component of 'pathname'
  */
  static char * getFileName( const char * pathname, char * filename,
					int maxlength);

 /** Extracts the directory path component from a pathname.
  *
  * @return Directory path component of 'pathname' (with ending '/')
	* or './' if directory path not present.
  */
  static char * getDirName( const char * pathname, char * dirname,
					int maxlength);

 /** List of files (not directories) in the specified directory */
	static vector<string> listFiles( const char * dirname);

 /** List of sub-directories in the specified directory */
	static vector<string> listDirs( const char * dirname);

	
 /** Defines the filesystem separator. */
	static const char fileSystemSeparator = '/';
	
 /** Defines the current directory pathname. */
	static const char * currentDirectory;

};

#endif
